/**************************************
****    CDROM.C  -  File reading   ****
**************************************/

//-- Include files -----------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <SDL.h>
#include <ctype.h>
#include <string.h>

#include "neocd.h"
#include "zfile.h"
#include "fileio.h"

/*-- Definitions -----------------------------------------------------------*/
#define	PRG_TYPE		0x00
#define FIX_TYPE		0x01
#define SPR_TYPE		0x02
#define Z80_TYPE		0x03
#define PCM_TYPE		0x04
#define PAT_TYPE		0x05
#define PAL_TYPE		0x06
#define OBJ_TYPE		0x07
#define AXX_TYPE		0x08
#define BACKUP_RAM		0x09
#define UNKNOWN_TYPE	0x0f

#ifndef min
#define min(a, b) ((a) < (b)) ? (a) : (b)
#endif

#define with_image()			m68000_read_memory_8(0x108000 + 0x7ddc)

/*-- Global Variables ----------------------------------------------------*/
int archive_mode = MODE_RAW;

//-- Private Variables -------------------------------------------------------
static void *iso;
static FILE *ofp = NULL;
static int file_size;
extern char neogeo_region;

//-- Private Function --------------------------------------------------------
static void cdrom_load_title(void);
static void neogeo_undecode_fix(u8 *mem, u32 offset, u32 length);
static void neogeo_apply_patch(u16 *src, int bank, u32 offset);

//----------------------------------------------------------------------------
int open_file(char *name, int opt)
{
	char fullname[512];
	if (file_size != -1) return file_size;
	switch(archive_mode) {
		case MODE_RAW:
			if(opt != 0) break;
			strcpy(fullname, game_dir);
			strcat(fullname, name);
			ofp = fopen(fullname, "rb");
			if(!ofp) break;
			fseek(ofp, 0, SEEK_END);
			file_size = ftell(ofp);
			fseek(ofp, 0, SEEK_SET);
			return file_size;
		case MODE_ZIP:
			if(opt != 0) break;
			if(zopen(name) == -1) break;
			file_size = zsize();
			return file_size;
		case MODE_ISO:
			if(opt > 1) break;
			if(iso == NULL) break;
			file_size = iso_find(iso, name, opt);
			if(file_size == -1) break;
			return file_size;
	}
	printf("Loading File Failed: %s\n", name);
	file_size = -1;
	ofp = NULL;
	return -1;
}

//----------------------------------------------------------------------------
int read_file(void *ptr, int size)
{
	if (file_size == -1) return 0;
	switch(archive_mode) {
		case MODE_RAW:
			return fread(ptr, 1, size, ofp);
			break;
		case MODE_ZIP:
			return zread(ptr, size);
			break;
		case MODE_ISO:
			return iso_read(iso, ptr, size);
			break;
	}
	return 0;
}

//----------------------------------------------------------------------------
void close_file()
{
	switch(archive_mode) {
		case MODE_RAW:
			if(ofp) fclose(ofp);
			break;
		case MODE_ZIP:
			zclose();
			break;
		case MODE_ISO:
			break;
	}
	ofp = NULL;
	file_size = -1;
}

//----------------------------------------------------------------------------
static void cdrom_load_title(void)
{
	char region_chr[3] = {'J', 'U', 'E'};
	char fname[16];
	int i, readed = 0;

	for (i = neogeo_region & 3; i >= 0; i--)
	{
		sprintf(fname, "TITLE_%c.SYS", region_chr[i]);

		if (open_file(fname, 0) != -1)
		{
			read_file(neogeo_paletteram16, 0x5A0);
			readed = read_file(memory_region_gfx2, 0x1680);
			close_file();
		}
	}

	if (readed == 0) return;

	swab((char *)neogeo_paletteram16, (char *)neogeo_paletteram16, 0x5A0);
	for(i=0; i<720; i++)
		video_palette[i] = video_clut16[neogeo_paletteram16[i]];
	neogeo_decode_spr(memory_region_gfx2, 0, readed);
	video_draw_title();

	memset(memory_region_gfx2, 0, 4194304);
	memset(video_spr_usage, 0, 32768);
}

/******************************************************************************
	Global variable
******************************************************************************/

int neogeo_loadscreen;
int neogeo_cdspeed_limit;
int neogeo_loadfinished;


/******************************************************************************
	Local struct/variable
******************************************************************************/

static struct filelist_t
{
	char name[16];
	u32  offset;
	int  bank;
	int  type;
	u32  length;
	u32  sectors;
} filelist[32];


static u8  *cdrom_cache;
static u32 total_sectors;
static u32 loaded_sectors;
static int total_files;

static int firsttime_update;


/******************************************************************************
	Prototype
******************************************************************************/



/******************************************************************************
	Local
******************************************************************************/

static int hextodec(char c)
{
	switch (tolower(c))
	{
	case '0': return 0;
	case '1': return 1;
	case '2': return 2;
	case '3': return 3;
	case '4': return 4;
	case '5': return 5;
	case '6': return 6;
	case '7': return 7;
	case '8': return 8;
	case '9': return 9;
	case 'a': return 10;
	case 'b': return 11;
	case 'c': return 12;
	case 'd': return 13;
	case 'e': return 14;
	case 'f': return 15;
	default: return 0;
	}
}


/*------------------------------------------------------
	Return file type from extension
------------------------------------------------------*/

static int get_filetype(const char *ext)
{
	int type = UNKNOWN_TYPE;

	if (strcasecmp(ext, "PRG") == 0) type = PRG_TYPE;
	if (strcasecmp(ext, "FIX") == 0) type = FIX_TYPE;
	if (strcasecmp(ext, "SPR") == 0) type = SPR_TYPE;
	if (strcasecmp(ext, "Z80") == 0) type = Z80_TYPE;
	if (strcasecmp(ext, "PCM") == 0) type = PCM_TYPE;
	if (strcasecmp(ext, "PAT") == 0) type = PAT_TYPE;
	if (ext[0] == 'A') type = AXX_TYPE;

	return type;
}


/*------------------------------------------------------
	Reset progress in loading screeen
------------------------------------------------------*/

static void init_loading_progress(void)
{
	if (neogeo_loadscreen && with_image())
	{
		m68000_write_memory_32(0x108000 + 0x7694, total_sectors);
		m68000_write_memory_32(0x108000 + 0x7690, 0);
	}
}


/*------------------------------------------------------
	Update progress in loading screen
------------------------------------------------------*/

static void update_loading_progress(void)
{
	if (neogeo_loadscreen && with_image())
	{
		static int draw = 0;
		u32 prev_progress, progress;

		prev_progress = m68000_read_memory_32(0x108000 + 0x7690);

		loaded_sectors++;
		progress = ((loaded_sectors * 0x8000) / total_sectors) << 8;

		if (progress != prev_progress)
		{
			m68000_write_memory_32(0x108000 + 0x7690, progress);

			if (progress >= 0x800000)
				m68000_write_memory_32(0x108000 + 0x7690, 0x800000);

			m68000_execute2(m68000_read_memory_32(0x11c80c), 0);

			m68000_write_memory_8(0x108000 + 0x7793, 0x00);
			m68000_execute2(0xc0c8b2, 0);

			draw = 1;
		}

		if (neogeo_ngh == NGH_ssideki3) {
			z80_exec(300);
			my_timer();
			//timer_update_subcpu();
		}
		draw = neogeo_loading_screenrefresh(firsttime_update, draw);
		firsttime_update = 0;
	}
}


/*------------------------------------------------------
	Uploading CD-ROM cache to memory
------------------------------------------------------*/

static void upload_file(int fileno, u32 offset, u32 length)
{
	struct filelist_t *file = &filelist[fileno];
	u32 base;

	switch (file->type)
	{
	case PRG_TYPE:
	case AXX_TYPE:
		swab(cdrom_cache, memory_region_cpu1 + file->offset + offset, length);
		break;

	case FIX_TYPE:
		if (neogeo_loadscreen && with_image())
		{
			u32 offset1, offset2, start, end;
			u32 length1, length2;

			start = (file->offset >> 1) + offset;
			end   = (start + length) - 1;

			if (start >= 0x6000)
			{
				offset1 = 0;
				length1 = 0;
				offset2 = start << 1;
				length2 = length;
			}
			else if (end < 0x6000)
			{
				offset1 = start << 1;
				length1 = length;
				offset2 = 0;
				length2 = 0;
			}
			else
			{
				offset1 = start << 1;
				length1 = 0x6000 - start;
				offset2 = 0x6000 << 1;
				length2 = length - length1;
			}

			if (length1)
			{
				memcpy(memory_region_cpu1 + 0x115e06 + (offset1 >> 1), cdrom_cache, length1);
			}

			if (length2)
			{
				memcpy(memory_region_gfx1 + (offset2 >> 1), cdrom_cache + length1, length2);
				neogeo_decode_fix(memory_region_gfx1, offset2, length2);
			}
		}
		else
		{
			base = file->offset >> 1;
			memcpy(memory_region_gfx1 + base + offset, cdrom_cache, length);
			neogeo_decode_fix(memory_region_gfx1, file->offset + (offset << 1), length);
		}
		break;

	case SPR_TYPE:
		base = (file->bank << 20) + file->offset;
		memcpy(memory_region_gfx2 + base + offset, cdrom_cache, length);
		neogeo_decode_spr(memory_region_gfx2, base + offset, length);
		break;

	case Z80_TYPE:
		base = file->offset >> 1;
		memcpy(memory_region_cpu2 + base + offset, cdrom_cache, length);
		break;

	case PAT_TYPE:
		swab(cdrom_cache, cdrom_cache, length);
		neogeo_apply_patch((u16 *)cdrom_cache, file->bank, file->offset);
		break;

	case PCM_TYPE:
		base = (file->bank << 19) + (file->offset >> 1);
		memcpy(memory_region_sound1 + base + offset, cdrom_cache, length);
		break;
	}
}


/*------------------------------------------------------
	Loading file to CD-ROM cache
------------------------------------------------------*/

static int load_file(int fileno)
{
	struct filelist_t *file = &filelist[fileno];
	int f;
	u32 length, next = 0;
	int size;

	if (file->type == Z80_TYPE)
		m68000_write_memory_8(0xff0183, 0);
	
	size = open_file(file->name, 0);

	if (size == -1)
	{
		//fatalerror(TEXT(COULD_NOT_OPEN_FILE), file->name);
		return -1;
	}
	printf("Loading File: %s(%d)\n", file->name, size);

	if (neogeo_loadscreen && with_image())
	{
		while (running && ((length = read_file(cdrom_cache, 0x800)) != 0))
		{
			upload_file(fileno, next, length);
			next += length;
			update_loading_progress();
		}
	}
	else
	{
		while (running && ((length = read_file(cdrom_cache, 0x800)) != 0))
		{
			upload_file(fileno, next, length);
			next += length;
		}
	}

	close_file();

	if (running && (file->type == Z80_TYPE))
		m68000_write_memory_8(0xff0183, 0xff);

	return 0;
}


/*--------------------------------------------------------
	Load screen data checking
--------------------------------------------------------*/

static u32 check_offset;

static int check_screen_data(u32 offset, int type)
{
	u32 data_offset;
	int data_type;

	check_offset = offset;

	do
	{
		data_type = m68000_read_memory_32(check_offset);
		check_offset += 4;

		data_offset = m68000_read_memory_32(check_offset);
		check_offset += 4;

		if (data_type == type + 1)
			return -1;

		if (data_type == -1)
			return -1;

	} while (data_type != type);

	return data_offset;
}


/*--------------------------------------------------------
	Uploading FIX sprite in loading screen
--------------------------------------------------------*/

static void loading_upload_fix(void)
{
	u8 *src, *dst;
	u32 offset;

	// FIX׫髤Ȫ
	src = memory_region_gfx1;
	dst = memory_region_cpu1 + 0x115e06;
	memcpy(dst, src, 0x6000);
	neogeo_undecode_fix(dst, 0, 0x6000);

	// ǫիȪΫ??FIX׫髤Ȫ?
	src = memory_region_user1 + 0x7c000;
	dst = memory_region_gfx1;
	memcpy(dst, src, 0x4000);
	neogeo_decode_fix(dst, 0, 0x4000);

	offset = 0x120002;

	// ??Ϋ??FIX׫髤Ȫ?
	while (1)
	{
		u16 fix_offs, size;

		offset = check_screen_data(offset, 1);
		if (offset == 0xffffffff)
			break;

		fix_offs = m68000_read_memory_32(offset);
		offset += 4;

		size = m68000_read_memory_32(offset);
		offset += 4;

		src = memory_region_cpu1;
		dst = memory_region_gfx1;
		memcpy(dst + (fix_offs >> 1), src + offset, size);
		neogeo_decode_fix(dst, fix_offs, size);

		offset = check_offset;
	}
}


/*--------------------------------------------------------
	Uploading palette in loading screen
--------------------------------------------------------*/

static void loading_upload_palette(void)
{
	int i;
	u32 src, dst, offset;

	// ΫѫëȪ
	src = 0x400000;
	dst = 0x11be06;

	for (i = 0; i < 0x200; i += 4)
	{
		m68000_write_memory_32(dst, m68000_read_memory_32(src));
		src += 4;
		dst += 4;
	}

	// BIOSΫѫëȪ?
	src = 0xc1701c;
	dst = 0x400000;

	for (i = 0; i < 0x200; i += 4)
	{
		m68000_write_memory_32(dst, m68000_read_memory_32(src));
		src += 4;
		dst += 4;
	}

	offset = 0x120002;

	// ??ΫѫëȪ?
	while (1)
	{
		u16 palno;

		src = check_screen_data(offset, 2);
		if (src == 0xffffffff)
			break;

		palno = m68000_read_memory_16(src);
		src += 2;

		dst = 0x400000 + (palno << 5);

		for (i = 0; i < 16; i++)
		{
			m68000_write_memory_16(dst, m68000_read_memory_16(src));
			src += 2;
			dst += 2;
		}

		offset = check_offset;
	}
}


/*--------------------------------------------------------
	Start loading screen
--------------------------------------------------------*/

static void loading_screen_start(void)
{
	if (neogeo_loadscreen)
	{
		firsttime_update = 1;

		if (with_image())
		{
			int i;
			u32 offset;

			//video_clear_screen();

			// Save FIX plane
			for (i = 0; i < 0x500; i++)
				m68000_write_memory_16(0x110804 + i * 2, neogeo_vidram16[0x7000 + i]);

			// Setup load screen
			m68000_execute2(m68000_read_memory_32(0x11c808), 0);

			loading_upload_fix();
			loading_upload_palette();

			offset = m68000_read_memory_32(0x11c80c);
			if (offset != 0xc0c814)
			{
				i = 64;	 // for safety

				if (NGH_NUMBER(0x0096))	// aof3
				{
					while (i--)
					{
						offset = m68000_read_memory_32(0x11c80c);
						m68000_execute2(offset, 0);

						if (offset == 0x1244f0 || offset == 0x124534
						||  offset == 0x1245b2 || offset == 0x1245ce)
							break;
					}
				}
				else if (NGH_NUMBER(0x0234))	// lastblad
				{
					while (i--)
					{
						offset = m68000_read_memory_32(0x11c80c);
						m68000_execute2(offset, 0);

						if (offset == 0x124550 || offset == 0x124b28)
							break;
					}
				}
				else if (NGH_NUMBER(0x0243))	// lastbld2
				{
					while (i--)
					{
						offset = m68000_read_memory_32(0x11c80c);
						m68000_execute2(offset, 0);

						if (offset == 0x1244aa || offset == 0x124d9c)
							break;
					}
				}
			}

			fix_disable_w(0);
			spr_disable_w(1);
			video_enable_w(1);
		}
		else if (NGH_NUMBER(0x0212))	// overtop
		{
			u8 *src, *dst;

			//video_clear_screen();

			src = memory_region_cpu1 + 0xe0000;
			dst = memory_region_gfx1;
			swab(src, dst, 0x20000);
			neogeo_decode_fix(dst, 0, 0x20000);

			m68000_write_memory_32(0x11c80c, 0x8db0);
		}
	}
}


/*--------------------------------------------------------
	Stop loading screen
--------------------------------------------------------*/

static void loading_screen_stop(void)
{
	if (neogeo_loadscreen)
	{
		if (with_image())
		{
			u8 *src, *dst;
			u32 src_offs, dst_offs;
			int i;

			// Restore palettes
			src_offs = 0x11be06;
			dst_offs = 0x400000;

			for (i = 0; i < 0x200; i += 4)
			{
				m68000_write_memory_32(dst_offs, m68000_read_memory_32(src_offs));
				src_offs += 4;
				dst_offs += 4;
			}

			// Restore FIX data
			src = memory_region_cpu1 + 0x115e06;
			dst = memory_region_gfx1;
			memcpy(dst, src, 0x6000);
			neogeo_decode_fix(dst, 0, 0x6000);

			// Restore FIX plane
			for (i = 0; i < 0x500; i++)
				neogeo_vidram16[0x7000 + i] = m68000_read_memory_16(0x110804 + i * 2);

			fix_disable_w(0);
			spr_disable_w(0);
			video_enable_w(1);
		}
		else if (NGH_NUMBER(0x0212))	// overtop
		{
			m68000_write_memory_32(0x11c80c, 0x8854);
		}
	}
}


/******************************************************************************
	CD-ROM Interface
******************************************************************************/

/*------------------------------------------------------
	CD-ROM Interface initialize
------------------------------------------------------*/

void cdrom_init(void)
{
	cdrom_cache = memory_region_cpu1 + 0x111204;
	neogeo_loadscreen = 0;
	neogeo_cdspeed_limit = 0;
	neogeo_loadfinished = 0;

	archive_mode = MODE_RAW;
	iso = iso_open(game_dir);
	if(iso != NULL) {
		archive_mode = MODE_ISO;
	} else if(zip_open(game_dir) != -1) {
		archive_mode = MODE_ZIP;
	}

	if(archive_mode == MODE_RAW) {
		char *p = strrchr(game_dir, '/');
		if(p == NULL) p = strrchr(game_dir, '\\');
		if(p != NULL) *p = 0;
#ifdef WIN32
		strcat(game_dir, "\\");
#else
		strcat(game_dir, "/");
#endif
	}

	printf("Archive: %s\n", game_dir);
	printf("Archive mode: %s\n", archive_mode == MODE_RAW ? "Raw" : archive_mode == MODE_ZIP ? "Zip" : "ISO");
	
	file_size = -1;
	ofp = NULL;
}


/*------------------------------------------------------
	CD-ROM Interface shutdown
------------------------------------------------------*/

void cdrom_shutdown(void)
{
	close_file();
	switch(archive_mode) {
		case MODE_RAW:
			break;
		case MODE_ZIP:
			zip_close();
			break;
		case MODE_ISO:
			iso_close(iso);
			iso = NULL;
			break;
	}
}


/*------------------------------------------------------
	IPL.TXT Process
------------------------------------------------------*/

int cdrom_process_ipl(void)
{
	struct filelist_t *file = &filelist[0];
    int i, j, f;
    u32 length;
    char linebuf[32], *buf, *p, *ext;
	char region_chr[3] = { 'J','U','E' };
	int cont, opt;

	//zip_open(game_dir);
	neogeo_loadfinished = 0;

	if (neogeo_loadscreen)
	{
		char fname[16];

		for (i = neogeo_region & 3; i >= 0; i--)
		{
			sprintf(fname, "LOGO_%c.PRG", region_chr[i]);

			if (open_file(fname, 0) != -1)
			{
				u8 *mem = memory_region_cpu1 + 0x120000;

				length = read_file(mem, 0x10000);
				close_file();

				swab(mem, mem, length);
			}
		}
	}
	else
	{
		now_loading();
	}
	
	cont = 1;opt = 0;
	while(cont) {
		memset(cdrom_cache, 0, 0x800);
	
	    if ((length = open_file("IPL.TXT", opt)) == -1)
	    {
			//zip_close();
			return 0;
		}

		read_file(cdrom_cache, length);
		close_file();
		
		total_sectors = 0;
		loaded_sectors = 0;
		total_files = 0;
	
		p = buf = (char *)cdrom_cache;
	
		while ((u32)p - (u32)cdrom_cache < length)
		{
			memset(linebuf, 0, 32);
	
			p = strchr(buf, 0x0a);
			if (p == NULL) break;
	
			*p++ = 0;
	
			strcpy(linebuf, buf);
	
			buf = p;
	
			if (strlen(linebuf) < 3) break;

			file->bank   = 0;
			file->offset = 0;
	
			i = 0;
			j = 0;
			while (linebuf[i] != ',')
			{
				file->name[j++] = toupper(linebuf[i++]);
			}
			file->name[j] = '\0';
	
			ext = &file->name[j - 3];
			if (ext[0] == 'O') strcpy(ext, "SPR");
			file->type = get_filetype(ext);
	
			i++;
			while (linebuf[i] != ',')
			{
				file->bank *= 10;
				file->bank += linebuf[i++] - '0';
			}
			file->bank &= 0x03;
	
			i++;
			while (linebuf[i] != 0x0d)
			{
				file->offset *= 16;
				file->offset += hextodec(linebuf[i++]);
			}
	
			file->length = open_file(file->name, 0);close_file();
			file->sectors = (file->length + 0x7ff) / 0x800;
			if (file->length != -1) cont = 0;
			else break;
	
			total_sectors += file->sectors;
			total_files++;
	
			file++;
		}
		opt++;
	}

	if (!neogeo_loadscreen) {
		cdrom_load_title();
	}

	m68000_write_memory_8(0x108000 + 0x7ddc, 0x01);

	init_loading_progress();
	loading_screen_start();

	for (i = 0; i < total_files; i++)
	{
		if (!running) return 1;
		load_file(i);
	}

	loading_screen_stop();

	m68000_write_memory_8(0x108000 + 0x7ddc, 0x00);
	m68000_write_memory_32(0x108000 + 0x7690, 0);

	//video_clear_screen();

	if (!neogeo_loadscreen) neogeo_loadfinished = 1;

	frameskip_reset();

	//zip_close();

	return 1;
}


/*------------------------------------------------------
	ի??
------------------------------------------------------*/

void cdrom_load_files(void)
{
	struct filelist_t *file = &filelist[0];
	u32 offset, src, dst, data, save1, save2;
	int i;

	neogeo_loadfinished = 0;

    src = m68000_get_reg(NULL,M68K_A0);

    if (m68000_read_memory_8(src) == 0)
		return;

	//zip_open(game_dir);

	if (!neogeo_loadscreen && with_image())
		now_loading();

	save1 = m68000_read_memory_8(0x108000 + 0x7d80);
	save2 = m68000_read_memory_8(0x108000 + 0x7dc2);

    cdda_stop();

	offset = 0x115a06;
	for (i = 0; i < 0x20; i++)
	{
		memset(memory_region_cpu1 + offset, 0, 0x20);
		m68000_write_memory_16(offset + 0x1c, 0xffff);
		offset += 0x20;
	}

	offset = 0x115a06;

	do
	{
		dst = offset;

		while ((data = m68000_read_memory_8(src++)) != 0)
		{
			if (data >= 'a') data -= 0x20;
			m68000_write_memory_8(dst++, data);
		}

		if (m68000_read_memory_8(dst - 3) == 'O')
		{
			m68000_write_memory_8(dst - 3, 'S');
			m68000_write_memory_8(dst - 2, 'P');
			m68000_write_memory_8(dst - 1, 'R');
		}

		dst = offset + 0x10;
		data = m68000_read_memory_8(src++) & 0x03;
		m68000_write_memory_8(dst++, data);

		src = (src + 1) & 0xfffffe;
		dst = (dst + 1) & 0xfffffe;

		m68000_write_memory_32(dst, m68000_read_memory_32(src));

		src += 4;
		offset += 0x20;

	} while (m68000_read_memory_8(src) != 0);

	offset = 0x115a06;

	total_sectors = 0;
	loaded_sectors = 0;
	total_files = 0;

	do
	{
		char *p;

		swab(memory_region_cpu1 + offset, file->name, 16);

		for (i = 0; file->name[i]; i++)
			if (!isprint(file->name[i]))
				break;

		if ((p = strchr(file->name, ';')) != NULL) *p = '\0';

		file->bank   = m68000_read_memory_8(offset + 0x10);
		file->offset = m68000_read_memory_32(offset + 0x12);
		file->type   = get_filetype(strrchr(file->name, '.') + 1);

		file->length = open_file(file->name, 0);close_file();
		total_sectors += (file->length + 0x7ff) / 0x800;
		total_files++;

		offset += 0x20;
		file++;

	} while (memory_region_cpu1[offset] != 0);

	init_loading_progress();
	loading_screen_start();

	for (i = 0; i < total_files; i++)
	{
		if (!running) return;
		load_file(i);
	}

	loading_screen_stop();

	//if (with_image() && !neogeo_loadscreen)
	//	video_clear_screen();

	m68000_write_memory_8(0x108000 + 0x7e88, 0x00);	// unknown, need to clear by '0'
	m68000_write_memory_8(0x108000 + 0x7ddd, 0x00);	// unknown, need to clear by '0'
	m68000_write_memory_8(0x108000 + 0x76db, 0x01);	// unknown, need to set '1'
	m68000_write_memory_8(0x108000 + 0x7ec4, 0x01);	// unknown, need to set '1'

	m68000_write_memory_32(0x10f742, 0);	// unknown, need to clear by '0'
	m68000_write_memory_32(0x10f746, 0);	// unknown, need to clear by '0'

	m68000_write_memory_8(0x108000 + 0x7657, 0x00);	// CD-ROM loading flag
	m68000_write_memory_8(0x108000 + 0x76d9, 0x01);	// Input enable (?)
	m68000_write_memory_8(0x108000 + 0x76c3, 0x01);	// CDDA enable
	m68000_write_memory_8(0x108000 + 0x76f6, 0xff);	// CDDA command
	m68000_write_memory_8(0x108000 + 0x764b, 0xff);	// CDDA track

	m68000_write_memory_8(0x108000 + 0x7ddc, 0x00);	// with/without loading screen
	m68000_write_memory_8(0x108000 + 0x7e85, 0xff);	// unknown

	m68000_write_memory_8(0x108000 + 0x7dc2, save2);
	m68000_write_memory_8(0x108000 + 0x7d80, save1);

	frameskip_reset();

	if (!neogeo_loadscreen) neogeo_loadfinished = 1;

	//zip_close();
}


/*------------------------------------------------------
	SPR sprite decoding
------------------------------------------------------*/

void neogeo_decode_spr(u8 *mem, u32 offset, u32 length)
{
	u32 tileno, numtiles = length / 128;
	u8 *base  = mem + offset;
	u8 *usage = video_spr_usage + (offset >> 7);

	for (tileno = 0; tileno < numtiles; tileno++)
	{
		u8 swap[128];
		u32 *gfxdata;
		int x,y;
		u32 pen;
		int opaque = 0;

		gfxdata = (u32*)&base[tileno << 7];

		memcpy(swap, gfxdata, 128);

		for (y = 0;y < 16;y++)
		{
			u32 dw, data;

			dw = 0;
			for (x = 0;x < 8;x++)
			{
				pen  = ((swap[64 + 4*y + 2] >> x) & 1) << 3;
				pen |= ((swap[64 + 4*y + 3] >> x) & 1) << 2;
				pen |= ((swap[64 + 4*y + 0] >> x) & 1) << 1;
				pen |= ((swap[64 + 4*y + 1] >> x) & 1) << 0;
				dw |= pen << 4*x;
			}

			*(gfxdata++) = dw;
			opaque |= dw;

			dw = 0;
			for (x = 0;x < 8;x++)
			{
				pen  = ((swap[4*y + 2] >> x) & 1) << 3;
				pen |= ((swap[4*y + 3] >> x) & 1) << 2;
				pen |= ((swap[4*y + 0] >> x) & 1) << 1;
				pen |= ((swap[4*y + 1] >> x) & 1) << 0;
				dw |= pen << 4*x;
			}

			*(gfxdata++) = dw;
			opaque |= dw;
		}

		if (opaque)
			*usage = 1;
		else
			*usage = 0;
		usage++;
	}
}


/*------------------------------------------------------
	FIX sprite decoding
------------------------------------------------------*/

#define decode_fix(n)				\
{									\
	tile = buf[n];					\
	*p++ = tile;					\
	opaque += (tile & 0x0f) != 0;	\
	opaque += (tile >> 4) != 0;		\
}

void neogeo_decode_fix(u8 *mem, u32 offset, u32 length)
{
	u32 i, j;
	u8 tile, opaque;
	u8 *p, buf[32];
	u8 *usage = &video_fix_usage[offset >> 6];
	u8 *base  = &mem[offset >> 1];

	for (i = 0; i < length; i += 32)
	{
		opaque  = 0;

		memcpy(buf, &base[i], 32);
		p = &base[i];

		for (j = 0; j < 8; j++)
		{
			decode_fix(j + 16);
			decode_fix(j + 24);
			decode_fix(j +  0);
			decode_fix(j +  8);
		}

		if (opaque)
			*usage = (opaque == 64) ? 1 : 2;
		else
			*usage = 0;
		*usage++;
	}
}


/*------------------------------------------------------
	FIX sprite recovery
------------------------------------------------------*/

#define undecode_fix(n)				\
{									\
	buf[n] = *p++;					\
}

static void neogeo_undecode_fix(u8 *mem, u32 offset, u32 length)
{
	u32 i, j;
	u8 *p, buf[32];
	u8 *base = &mem[offset >> 1];

	for (i = 0; i < length; i += 32)
	{
		memcpy(buf, &base[i], 32);
		p = &base[i];

		for (j = 0; j < 8; j++)
		{
			undecode_fix(j + 16);
			undecode_fix(j + 24);
			undecode_fix(j +  0);
			undecode_fix(j +  8);
		}
	}
}


/*------------------------------------------------------
	Z80 patch apply
------------------------------------------------------*/

#define PATCH_Z80(a, b)					\
{										\
	dst[((a) + 0)] =  (b)       & 0xff;	\
	dst[((a) + 1)] = ((b) >> 8) & 0xff;	\
}

static void neogeo_apply_patch(u16 *src, int bank, u32 offset)
{
	u8 *dst = memory_region_cpu2;

	offset = (((bank * 0x100000) + offset) >> 8) & 0xffff;

	while (*src)
	{
		PATCH_Z80(src[0] + 0,  (src[1] + offset) >> 1);
		PATCH_Z80(src[0] + 2, ((src[2] + offset) >> 1) - 1);

		if (src[3] && src[4])
		{
			PATCH_Z80(src[0] + 5,  (src[3] + offset) >> 1);
			PATCH_Z80(src[0] + 7, ((src[4] + offset) >> 1) - 1);
		}

		src += 5;
	}
}
