/*
    Copyright (C) 2000 Herbert Valerio Riedel <hvr@gnu.org>
    Copyright (C) 2003, 2004, 2005, 2006 Rocky Bernstein <rocky@gnu.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include <errno.h>
#include <stdio.h>

#include <time.h>
#include <stdint.h>
#include <stdbool.h>
#include <inttypes.h>
#include <sys/types.h>

typedef uint8_t  iso711_t;
typedef int8_t   iso712_t;
typedef uint16_t iso721_t;
typedef uint16_t iso722_t;
typedef uint32_t iso723_t;
typedef uint32_t iso731_t;
typedef uint32_t iso732_t;
typedef uint64_t iso733_t;

typedef char     achar_t;
typedef char     dchar_t;

#ifndef EMPTY_ARRAY_SIZE
#define EMPTY_ARRAY_SIZE 0
#endif

#if __GNUC__ > 2 || (__GNUC__ == 2 && __GNUC_MINOR__ > 4)
# define GNUC_PACKED __attribute__((packed))
#else
# define GNUC_PACKED
#endif

#if defined(__GNUC__)
  /* for GCC we try to use GNUC_PACKED */
# define PRAGMA_BEGIN_PACKED
# define PRAGMA_END_PACKED
#elif defined(HAVE_ISOC99_PRAGMA)
  /* should work with most EDG-frontend based compilers */
# define PRAGMA_BEGIN_PACKED _Pragma("pack(1)")
# define PRAGMA_END_PACKED   _Pragma("pack()")
#else /* neither gcc nor _Pragma() available... */
  /* ...so let's be naive and hope the regression testsuite is run... */
# define PRAGMA_BEGIN_PACKED
# define PRAGMA_END_PACKED
#endif

typedef int32_t lsn_t;
  
extern enum iso_enum1_s {
  ISO_PVD_SECTOR      =   16,
  ISO_EVD_SECTOR      =   17,
  LEN_ISONAME         =   31,
  ISO_MAX_SYSTEM_ID   =   32,
  MAX_ISONAME         =   37,
  ISO_MAX_PREPARER_ID =  128,
  MAX_ISOPATHNAME     =  255,
  ISO_BLOCKSIZE       = 2048 
} iso_enums1;

#define ISO_MAX_PUBLISHER_ID 128
#define ISO_MAX_APPLICATION_ID 128
#define ISO_MAX_VOLUME_ID 32
#define ISO_MAX_VOLUMESET_ID 128
#define ISO_STANDARD_ID "CD001" 

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

PRAGMA_BEGIN_PACKED

#define	_delta(from, to)	((to) - (from) + 1)

struct	iso9660_ltime_s {
  char	 lt_year	[_delta(   1,	4)];
  char	 lt_month	[_delta(   5,	6)];
  char	 lt_day		[_delta(   7,	8)];
  char	 lt_hour	[_delta(   9,	10)];
  char	 lt_minute	[_delta(  11,	12)];
  char	 lt_second	[_delta(  13,	14)];
  char	 lt_hsecond	[_delta(  15,	16)];
  iso712_t lt_gmtoff;
} GNUC_PACKED;

struct	iso9660_dtime_s {
  iso711_t 	dt_year;
  iso711_t 	dt_month;
  iso711_t	dt_day;
  iso711_t	dt_hour;
  iso711_t	dt_minute;
  iso711_t	dt_second;
  iso712_t	dt_gmtoff;
} GNUC_PACKED;

typedef struct iso9660_ltime_s  iso9660_ltime_t;
typedef struct iso9660_dtime_s  iso9660_dtime_t;

struct iso9660_dir_s {
  iso711_t         length;
  iso711_t         xa_length;
  iso733_t         extent;
  iso733_t         size;
  iso9660_dtime_t  recording_time;
  uint8_t          file_flags;
  iso711_t         file_unit_size;
  iso711_t         interleave_gap;
  iso723_t         volume_sequence_number;
  iso711_t         filename_len;
  char             filename[EMPTY_ARRAY_SIZE];
} GNUC_PACKED;

typedef struct iso9660_dir_s    iso9660_dir_t;

struct iso9660_pvd_s {
  iso711_t         type;
  char             id[5];
  iso711_t         version;
  char             unused1[1];
  achar_t          system_id[ISO_MAX_SYSTEM_ID];
  dchar_t          volume_id[ISO_MAX_VOLUME_ID];
  uint8_t          unused2[8];
  iso733_t         volume_space_size;
  uint8_t          unused3[32];
  iso723_t         volume_set_size;
  iso723_t         volume_sequence_number;
  iso723_t         logical_block_size;
  iso733_t         path_table_size;
  iso731_t         type_l_path_table;
  iso731_t         opt_type_l_path_table;
  iso732_t         type_m_path_table;
  iso732_t         opt_type_m_path_table;
  iso9660_dir_t    root_directory_record;
  char             root_directory_filename;
  dchar_t          volume_set_id[ISO_MAX_VOLUMESET_ID];
  achar_t          publisher_id[ISO_MAX_PUBLISHER_ID];
  achar_t          preparer_id[ISO_MAX_PREPARER_ID];
  achar_t          application_id[ISO_MAX_APPLICATION_ID];
  dchar_t          copyright_file_id[37];
  dchar_t          abstract_file_id[37];
  dchar_t          bibliographic_file_id[37];
  iso9660_ltime_t  creation_date;
  iso9660_ltime_t  modification_date;
  iso9660_ltime_t  expiration_date;
  iso9660_ltime_t  effective_date;
  iso711_t         file_structure_version;
  uint8_t          unused4[1];
  char             application_data[512];
  uint8_t          unused5[653];
} GNUC_PACKED;

typedef struct iso9660_pvd_s  iso9660_pvd_t;

PRAGMA_END_PACKED

struct _iso9660_s {
  FILE *stream;
  iso9660_pvd_t pvd;
  uint8_t *root_dir;
  uint32_t root_size;
  lsn_t find_lsn;
  uint32_t find_size;
  uint32_t buf_size;
  uint32_t buf_pos;
  int mode;
  char find_buf[2048];
};

typedef struct _iso9660_s iso9660_t; 

#ifdef __cplusplus
}
#endif /* __cplusplus */

#define MAX_DEPTH 32
#define MAX_PATH 2048

#ifndef UINT32_C
# if defined (SIZEOF_INT) && SIZEOF_INT == 4
#  define UINT32_C(c) c ## U
# elif defined (SIZEOF_LONG) && SIZEOF_LONG == 4
#  define UINT32_C(c) c ## UL
# else
#  define UINT32_C(c) c ## U
# endif
#endif
  
#ifndef UINT64_C
# if defined (SIZEOF_LONG) && SIZEOF_LONG == 8
#  define UINT64_C(c) c ## UL
# elif defined (SIZEOF_INT) && SIZEOF_INT == 8
#  define UINT64_C(c) c ## U
# else
#  define UINT64_C(c) c ## ULL
# endif
#endif

/**! ISO 9660 directory flags. */
extern enum iso_flag_enum_s {
  ISO_FILE	      =   0,
  ISO_EXISTENCE	  =   1,
  ISO_DIRECTORY	  =   2,
  ISO_ASSOCIATED  =   4,
  ISO_RECORD	  =   8,
  ISO_PROTECTION  =  16,
  ISO_DRESERVED1  =  32,
  ISO_DRESERVED2  =  64,
  ISO_MULTIEXTENT = 128,
} iso_flag_enums;

/**! Volume descriptor types */
extern enum iso_vd_enum_s {
  ISO_VD_BOOT_RECORD   =  0,
  ISO_VD_PRIMARY       =  1,
  ISO_VD_SUPPLEMENTARY =  2,
  ISO_VD_PARITION      =  3,
  ISO_VD_END	       = 255
} iso_vd_enums;

  
/** 16-bit big-endian to little-endian */
#define UINT16_SWAP_LE_BE_C(val) ((uint16_t) ( \
    (((uint16_t) (val) & (uint16_t) 0x00ffU) << 8) | \
    (((uint16_t) (val) & (uint16_t) 0xff00U) >> 8)))

/** 32-bit big-endian to little-endian */
#define UINT32_SWAP_LE_BE_C(val) ((uint32_t) ( \
    (((uint32_t) (val) & (uint32_t) 0x000000ffU) << 24) | \
    (((uint32_t) (val) & (uint32_t) 0x0000ff00U) <<  8) | \
    (((uint32_t) (val) & (uint32_t) 0x00ff0000U) >>  8) | \
    (((uint32_t) (val) & (uint32_t) 0xff000000U) >> 24)))

/** 64-bit big-endian to little-endian */
#define UINT64_SWAP_LE_BE_C(val) ((uint64_t) ( \
    (((uint64_t) (val) & (uint64_t) UINT64_C(0x00000000000000ff)) << 56) | \
    (((uint64_t) (val) & (uint64_t) UINT64_C(0x000000000000ff00)) << 40) | \
    (((uint64_t) (val) & (uint64_t) UINT64_C(0x0000000000ff0000)) << 24) | \
    (((uint64_t) (val) & (uint64_t) UINT64_C(0x00000000ff000000)) <<  8) | \
    (((uint64_t) (val) & (uint64_t) UINT64_C(0x000000ff00000000)) >>  8) | \
    (((uint64_t) (val) & (uint64_t) UINT64_C(0x0000ff0000000000)) >> 24) | \
    (((uint64_t) (val) & (uint64_t) UINT64_C(0x00ff000000000000)) >> 40) | \
    (((uint64_t) (val) & (uint64_t) UINT64_C(0xff00000000000000)) >> 56)))

#ifndef UINT16_SWAP_LE_BE
# define UINT16_SWAP_LE_BE UINT16_SWAP_LE_BE_C
#endif

#ifndef UINT32_SWAP_LE_BE
# define UINT32_SWAP_LE_BE UINT32_SWAP_LE_BE_C
#endif

#ifndef UINT64_SWAP_LE_BE
# define UINT64_SWAP_LE_BE UINT64_SWAP_LE_BE_C
#endif

inline static 
uint16_t uint16_swap_le_be (const uint16_t val)
{
  return UINT16_SWAP_LE_BE (val);
}

inline static 
uint32_t uint32_swap_le_be (const uint32_t val)
{
  return UINT32_SWAP_LE_BE (val);
}

inline static 
uint64_t uint64_swap_le_be (const uint64_t val)
{
  return UINT64_SWAP_LE_BE (val);
}

# define UINT8_TO_BE(val)      ((uint8_t) (val))
# define UINT8_TO_LE(val)      ((uint8_t) (val))
#ifdef WORDS_BIGENDIAN
# define UINT16_TO_BE(val)     ((uint16_t) (val))
# define UINT16_TO_LE(val)     ((uint16_t) UINT16_SWAP_LE_BE(val))

# define UINT32_TO_BE(val)     ((uint32_t) (val))
# define UINT32_TO_LE(val)     ((uint32_t) UINT32_SWAP_LE_BE(val))

# define UINT64_TO_BE(val)     ((uint64_t) (val))
# define UINT64_TO_LE(val)     ((uint64_t) UINT64_SWAP_LE_BE(val))
#else
# define UINT16_TO_BE(val)     ((uint16_t) UINT16_SWAP_LE_BE(val))
# define UINT16_TO_LE(val)     ((uint16_t) (val))

# define UINT32_TO_BE(val)     ((uint32_t) UINT32_SWAP_LE_BE(val))
# define UINT32_TO_LE(val)     ((uint32_t) (val))

# define UINT64_TO_BE(val)     ((uint64_t) UINT64_SWAP_LE_BE(val))
# define UINT64_TO_LE(val)     ((uint64_t) (val))
#endif

/** symmetric conversions */
#define UINT8_FROM_BE(val)     (UINT8_TO_BE (val))
#define UINT8_FROM_LE(val)     (UINT8_TO_LE (val))
#define UINT16_FROM_BE(val)    (UINT16_TO_BE (val))
#define UINT16_FROM_LE(val)    (UINT16_TO_LE (val))
#define UINT32_FROM_BE(val)    (UINT32_TO_BE (val))
#define UINT32_FROM_LE(val)    (UINT32_TO_LE (val))
#define UINT64_FROM_BE(val)    (UINT64_TO_BE (val))
#define UINT64_FROM_LE(val)    (UINT64_TO_LE (val))

/** converter function template */
#define CVT_TO_FUNC(bits) \
 static inline uint ## bits ## _t \
 uint ## bits ## _to_be (uint ## bits ## _t val) \
 { return UINT ## bits ## _TO_BE (val); } \
 static inline uint ## bits ## _t \
 uint ## bits ## _to_le (uint ## bits ## _t val) \
 { return UINT ## bits ## _TO_LE (val); } \

CVT_TO_FUNC(8)
CVT_TO_FUNC(16)
CVT_TO_FUNC(32)
CVT_TO_FUNC(64)

#undef CVT_TO_FUNC

#define uint8_from_be(val)     (uint8_to_be (val))
#define uint8_from_le(val)     (uint8_to_le (val))
#define uint16_from_be(val)    (uint16_to_be (val))
#define uint16_from_le(val)    (uint16_to_le (val))
#define uint32_from_be(val)    (uint32_to_be (val))
#define uint32_from_le(val)    (uint32_to_le (val))
#define uint64_from_be(val)    (uint64_to_be (val))
#define uint64_from_le(val)    (uint64_to_le (val))

#define to_711(i)   uint8_to_le(i)
#define from_711(i) uint8_from_le(i)
#define to_721(i)   uint16_to_le(i)
#define from_721(i) uint16_from_le(i)
#define to_722(i)   uint16_to_be(i)
#define from_722(i) uint16_from_be(i)

static inline uint32_t
to_723(uint16_t i)
{
  return uint32_swap_le_be(i) | i;
}

static inline uint16_t 
from_723 (uint32_t p)
{
  if (uint32_swap_le_be (p) != p)
    printf ("from_723: broken byte order\n");

  return (0xFFFF & p);
}

#define to_731(i)   uint32_to_le(i)
#define from_731(i) uint32_from_le(i)
#define to_732(i)   uint32_to_be(i)
#define from_732(i) uint32_from_be(i)

static inline uint64_t to_733(uint32_t i) {
  return uint64_swap_le_be(i) | i;
}

static inline uint32_t from_733 (uint64_t p) {
  if (uint64_swap_le_be (p) != p)
    printf ("from_733: broken byte order\n");

  return (UINT32_C(0xFFFFFFFF) & p);
}

static inline uint32_t len2blocks (uint32_t i_len, uint16_t i_blocksize)
{
	uint32_t i_blocks;

	i_blocks = i_len / (uint32_t) i_blocksize;
	if (i_len % i_blocksize)
		i_blocks++;

	return i_blocks;
}

enum {
	MODE_ISO_2048 = 0,
	MODE_ISO_2352,
	MODE_1_2352,
	MODE_2_2352,
	MODE_2_2336,
};

long int iso_seek_read (const iso9660_t *p_iso, void *ptr, lsn_t start, long int size)
{
	long int ret, i;
	long int i_byte_offset;
	char *read_ptr = ptr;

	if (!p_iso) return 0;
	if ((ptr == NULL) || (size == 0)) return 0;
	
	for (i=0; i<size; i++) {
		switch(p_iso->mode) {
			case MODE_ISO_2048:
				i_byte_offset = (start + i) * 2048;
				break;
			case MODE_ISO_2352: {
				int page = start + i;
				int cur = page % 27;
				page -= cur;
				i_byte_offset = (page * 2352) + (cur * 2048);
				break;
			case MODE_1_2352:
				i_byte_offset = (start + i) * 2352 + 16;
				break;
			case MODE_2_2352:
				i_byte_offset = (start + i) * 2352 + 24;
				break;
			case MODE_2_2336:
				i_byte_offset = (start + i) * 2336 + 8;
				break;
			}
		}

		if (fseek (p_iso->stream, i_byte_offset, SEEK_SET) != 0) return 0;
		if (fread (read_ptr, ISO_BLOCKSIZE, 1, p_iso->stream) != 1) return 0;
		read_ptr += 2048;
	}

	return (int)read_ptr - (int)ptr;
}

static bool check_pvd (const iso9660_pvd_t *p_pvd)
{
	if ( ISO_VD_PRIMARY != from_711(p_pvd->type) ) {
		//fprintf (stderr, "unexpected PVD type %d\n", p_pvd->type);
		return false;
	}

	if (strncmp (p_pvd->id, ISO_STANDARD_ID, strlen (ISO_STANDARD_ID)))
	{
		//fprintf (stderr, "unexpected ID encountered (expected `"ISO_STANDARD_ID "', got `%.5s')\n", p_pvd->id);
		return false;
	}
	return true;
}

bool iso_read_superblock (iso9660_t *p_iso)
{
	if (0 == iso_seek_read (p_iso, &(p_iso->pvd), ISO_PVD_SECTOR, 1)) {
		//fprintf ( stderr, "error reading PVD sector (%d)\n", ISO_PVD_SECTOR );
		return false;
	}
	return check_pvd(&(p_iso->pvd));
}

void *iso_open (const char *psz_path)
{
    const char SYNC_HEADER[12] = { 0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0 };
    char  headbuf[16];
	iso9660_t *p_iso = (iso9660_t *) calloc(1, sizeof(iso9660_t)) ;
	iso9660_dir_t *p_dir;
	bool b_have_superblock;
	lsn_t lsn;
	uint32_t size, secsize;
    int ret;

	if (!p_iso) return NULL;

    p_iso->mode = MODE_ISO_2048;
	p_iso->stream = fopen( psz_path, "rb" );
	if (NULL == p_iso->stream) goto error;

	if (fread(headbuf, 1, 16, p_iso->stream) != 16) goto error;
    if (memcmp(SYNC_HEADER, headbuf, 12) == 0) {
		if(headbuf[15] == 1)
			p_iso->mode = MODE_1_2352;
		else if(headbuf[15] == 2)
			p_iso->mode = MODE_2_2352;
		else
			goto error;
	}

	b_have_superblock = iso_read_superblock(p_iso) ;
	if (!b_have_superblock && (p_iso->mode == MODE_ISO_2048)) {
		p_iso->mode = MODE_2_2336;
		b_have_superblock = iso_read_superblock(p_iso);
	}
	if (!b_have_superblock) goto error;
	
	printf("ISO mode: %d\n", p_iso->mode);

	p_dir = &(p_iso->pvd.root_directory_record);
	if ((p_dir->file_flags & ISO_DIRECTORY) == 0) goto error;

	lsn = from_733 (p_dir->extent);
	size = from_733 (p_dir->size);
	secsize = len2blocks (size, ISO_BLOCKSIZE);

	if (size != ISO_BLOCKSIZE * secsize)
		printf ("bad size for ISO9660 directory (%ud) should be (%lu)!\n", (unsigned)size, (unsigned long int) ISO_BLOCKSIZE * secsize);
		
	p_iso->root_dir = calloc(1, secsize * ISO_BLOCKSIZE);
	if(!p_iso->root_dir) goto error;

	ret = iso_seek_read(p_iso, p_iso->root_dir, lsn, secsize);
	if (ret != ISO_BLOCKSIZE * secsize) {
        free(p_iso->root_dir);
        goto error;
    }
    
    p_iso->root_size = ISO_BLOCKSIZE * secsize;
    p_iso->find_lsn = -1;

	return p_iso;

error:
	fclose(p_iso->stream);
	free(p_iso);
	return NULL;
}

void iso_close (void *piso)
{
	iso9660_t *p_iso = piso;
	if (NULL != p_iso) {
        free(p_iso->root_dir);
		fclose(p_iso->stream);
		free(p_iso);
	}
}

int iso_find (void *piso, const char *psz_path, int opt)
{
	iso9660_t *p_iso = piso;
	char *splitpath[MAX_DEPTH];
	char path[MAX_PATH];
	char *p, *tok;
	int depth = 0;
	int find;
	bool isfind, isdir;
	uint8_t *_dirbuf = NULL;
	iso9660_dir_t *p_dir;
	lsn_t lsn;
	uint32_t size;

	if(!p_iso || !psz_path) return 0;
	if(opt == 1) {
		if(p_iso->mode != MODE_ISO_2048) return 0;
		p_iso->mode = MODE_ISO_2352;
	}

	strcpy(path, psz_path);
	tok = strtok(path, "/\\");
	while(tok != NULL) {
		splitpath[depth++] = tok;
		tok = strtok(NULL, "/\\");
		if(depth > (MAX_DEPTH-2)) return 0;
	}
	if(depth == 0) return 0;

	_dirbuf = p_iso->root_dir;
	size = p_iso->root_size;	

	find = 0;isfind = true;
	while(isfind && (find < depth)) {
		unsigned offset;
		char name[MAX_PATH];
		int ret, namelen, pathlen;
		uint32_t secsize;
		
		offset = 0;isfind = false;pathlen = strlen(splitpath[find]);
		while(offset < size) {
			p_dir = (void *) &_dirbuf[offset];
			if(!p_dir->length) {
				offset++;
				continue;
			}

			namelen = from_711(p_dir->filename_len);
			if ('\0' == p_dir->filename[0] && 1 == namelen) {
				strcpy (name, ".");
				namelen = 1;
			} else if ('\1' == p_dir->filename[0] && 1 == namelen) {
				strcpy (name, "..");
				namelen = 2;
			} else {
				strncpy (name, p_dir->filename, namelen);
				if((namelen > pathlen) && (name[pathlen] == ';'))
					namelen = pathlen;
				name[namelen] = 0;
			}

			if((namelen == pathlen) && (strcasecmp(name, splitpath[find]) == 0)) {
				isfind = true;
				break;
			}

			offset += p_dir->length;
		}
		if(!isfind) {
			if(_dirbuf != p_iso->root_dir) free(_dirbuf);
			break;
		}

		find++;
		isdir = p_dir->file_flags & ISO_DIRECTORY;
		lsn = from_733 (p_dir->extent);
		size = from_733 (p_dir->size);
		secsize = len2blocks (size, ISO_BLOCKSIZE);

		if(_dirbuf != p_iso->root_dir) free(_dirbuf);
		if(!isdir || (find == depth)) break;

		if (size != ISO_BLOCKSIZE * secsize)
			printf ("bad size for ISO9660 directory (%ud) should be (%lu)!\n", (unsigned)size, (unsigned long int) ISO_BLOCKSIZE * secsize);
		size = ISO_BLOCKSIZE * secsize;

		_dirbuf = calloc(1, size);
		ret = iso_seek_read(p_iso, _dirbuf, lsn, secsize);
		if (ret != size) {
            free(_dirbuf);
            break;
        }

        lsn = 0;
	}
	
	if(isfind && !isdir && (find == depth)) {
		p_iso->find_lsn = lsn;
        p_iso->find_size = size;
        p_iso->buf_size = 0;
        p_iso->buf_pos = 0;
        return size;
    }

	p_iso->find_lsn = -1;
	return -1;
}

int iso_read(void *piso, void *ptr, int size)
{
	iso9660_t *p_iso = piso;
	char *read_ptr = ptr;
	int read_size = 0;

	if (p_iso == NULL || ptr == NULL || size == 0) return 0;
	if (size > p_iso->find_size) size = p_iso->find_size;

	while (read_size < size) {
		if (p_iso->buf_size > 0) {
			int get_size = p_iso->buf_size < size ? p_iso->buf_size : size;
			memcpy(read_ptr + read_size, p_iso->find_buf + p_iso->buf_pos, get_size);
			p_iso->buf_pos += get_size;
			p_iso->buf_size -= get_size;
			p_iso->find_size -= get_size;
			read_size += get_size;
		} else {
			p_iso->buf_size = iso_seek_read(p_iso, p_iso->find_buf, p_iso->find_lsn, 1);
			if(p_iso->buf_size = 2048) {
				p_iso->find_lsn++;
				p_iso->buf_pos = 0;
			} else {
				return read_size;
			}
/*
			int block = (size - read_size) / 2048;
			if(block > 0) {
				int get_size = iso_seek_read(p_iso, read_ptr + read_size, p_iso->find_lsn, block);
				read_size += get_size;
				if(block != (get_size / 2048)) return read_size;
				p_iso->find_lsn += block;
				p_iso->find_size -= get_size;
			} else {
				p_iso->buf_size = iso_seek_read(p_iso, p_iso->find_buf, p_iso->find_lsn, 1);
				if(p_iso->buf_size = 2048) {
					p_iso->find_lsn++;
					p_iso->buf_pos = 0;
				} else {
					return read_size;
				}
			}
*/
		}
	}

	return read_size;
}
