/**************************************
****  NEOCD.C  - Main Source File  ****
**************************************/

//-- Include Files -----------------------------------------------------------

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#ifdef WIN32
#include <windows.h>
#endif

#include <SDL.h>
#include <SDL_image.h>

#include "neocd.h"
#include "icon.h"
#include "filer.h"
#ifdef GP2X
#include "usbjoy.h"
#endif

#define REGION_JAPAN  0
#define REGION_USA    1
#define REGION_EUROPE 2

#define REGION REGION_JAPAN

#define msg_printf printf
#define BTN(x) (1 << x)

//-- Global Variables --------------------------------------------------------
char neogeo_region = REGION;
char config_game_name[80];
int scanline;

char game_name[16];
char game_dir[MAX_PATH];

char skip_next_frame;

char LastPath[256];
char launchDir[MAX_PATH];

int running;
int filer_option;

//-- Local Variables --------------------------------------------------------
static SDL_Surface *loading = NULL;
static int skin;
static int loading_skin = -1;
static int sound_enable = 0;
static int joy_count = 0;

static int cpu_reset_flag;
static Uint32 neocd_time; /* next time marker */

static char save_path[MAX_PATH];
static char config_file[MAX_PATH];
static char neocd_wm_title[255]=VERSION1;

// startup.bin data
#include "startup.h"

//-- Function Prototypes -----------------------------------------------------
int neogeo_init(void);
void neogeo_reset(void);
void neogeo_shutdown(void);
void neogeo_run(void);

void neogeo_read_gamename(void);
TICKER get_ticks();

//-- Platform Dependency -----------------------------------------------------
#ifdef GP2X
enum {
	GPX_VK_UP = 0,
	GPX_VK_UP_LEFT,
	GPX_VK_LEFT,
	GPX_VK_DOWN_LEFT,
	GPX_VK_DOWN,
	GPX_VK_DOWN_RIGHT,
	GPX_VK_RIGHT,
	GPX_VK_UP_RIGHT,

	GPX_VK_START,
	GPX_VK_SELECT,
	GPX_VK_FL,
	GPX_VK_FR,
	GPX_VK_FA,
	GPX_VK_FB,
	GPX_VK_FX,
	GPX_VK_FY,

	GPX_VK_VOL_UP,
	GPX_VK_VOL_DOWN,
	GPX_VK_TAT,

	GPX_VK_MAX
};

enum {
	JOY_UP = 0,
	JOY_DOWN,
	JOY_LEFT,
	JOY_RIGHT,

	JOY_BUTTON,
	JOY_BUTTON1,

	JOY_MAX = 37
};

enum {
	JOY_BASIC = 0,
	JOY_GREEN,
	JOY_DRAGON
};

#define GPX_CROSS_MASK ( BTN(GPX_VK_UP) | BTN(GPX_VK_DOWN) | BTN(GPX_VK_LEFT) | BTN(GPX_VK_RIGHT) )
#define GPX_DIAG_MASK ( BTN(GPX_VK_UP_LEFT) | BTN(GPX_VK_DOWN_LEFT) | BTN(GPX_VK_UP_RIGHT) | BTN(GPX_VK_DOWN_RIGHT) )

static struct joy_info {
	const char *name;
	const int type;
} joy_list[] = {
	{"GreenAsia Inc.    USB Joystick     ", JOY_GREEN},
	{"DragonRise Inc.   Generic   USB  Joystick  ", JOY_DRAGON},
	{NULL, JOY_BASIC}
};

static int joy_fd = -1;
static int mem_fd = -1;
volatile static u16 *gp2xregs = NULL;
static struct usbjoy *joys[4];
static int joy_type[4] = {0, };
static int joy_map[4][JOY_MAX];
static int input_map[GPX_VK_MAX];

char option_frontend[MAX_PATH] = "/usr/gp2x/gp2xmenu";
#else
static LARGE_INTEGER freq, base;
#endif

static Uint32 next_time;
static int frame_count;
static int draw_count;
static const Uint32 ONESEC = 1000000;

//----------------------------------------------------------------------------
void now_loading()
{
	if (loading_skin != skin) {
		char img_path[_MAX_PATH];
		sprintf(img_path, "%sskin%d/loading.png", launchDir, skin);
		if(loading)
			SDL_FreeSurface(loading);
		loading=IMG_Load(img_path);
		loading_skin=skin;
	}
	if (loading || game_title) {
        if(game_title) {
			SDL_Rect dstrect;
			dstrect.x = 320 - 144 - 16;
			dstrect.y = 240 - 80 - 16;
			SDL_BlitSurface(game_title, NULL, screen, &dstrect);
        } else if( (loading->w < 320) || (loading->h < 240) ) {
			SDL_Rect dstrect;
			dstrect.w = loading->w;
			dstrect.h = loading->h;
			if(dstrect.w < 320)
				dstrect.x = (320 - dstrect.w) >> 1;
			else
				dstrect.x = 0;
			if(dstrect.h < 240)
				dstrect.y = (240 - dstrect.h) >> 1;
			else
				dstrect.y = 0;
			SDL_BlitSurface(loading, NULL, screen, &dstrect);
		} else {
			SDL_BlitSurface(loading, NULL, screen, NULL);
		}
		SDL_Flip(screen);
	}
}

//----------------------------------------------------------------------------
int main(int argc, char* argv[])
{
	FILE *fp;
	SDL_RWops *iconfp;
	SDL_Surface *icon;
	int result;
	int i;

	// Register exit procedure
	atexit(neogeo_shutdown);

	// Initialise SDL
	if((SDL_Init(SDL_INIT_VIDEO|SDL_INIT_TIMER|SDL_INIT_JOYSTICK|SDL_INIT_NOPARACHUTE)==-1)) {
		printf("Could not initialize SDL: %s.\n", SDL_GetError());
		exit(-1);
	}

	//Hide Mouse Pointer
	SDL_ShowCursor(SDL_DISABLE);

#ifdef GP2X
	if ((screen=SDL_SetVideoMode(320, 240, 16, SDL_FULLSCREEN))==NULL)
#else
	if ((screen=SDL_SetVideoMode(320, 240, 16, /*SDL_FULLSCREEN|SDL_HWPALETTE|*/SDL_HWSURFACE))==NULL)
#endif
	{
		fprintf( stderr, "Could not set video mode: %s\n", SDL_GetError() );
		exit( -1 );
	}

	SDL_ShowCursor(SDL_DISABLE);

#ifdef GP2X
	joy_fd = open("/dev/GPIO", O_RDWR | O_NDELAY);
	if(joy_fd < 0) {
		fprintf (stderr, "Couldn't open /dev/GPIO device.\n");
		exit (1);
	}
	joy_count = 0;
#endif

#ifdef WIN32
	QueryPerformanceFrequency(&freq);
	QueryPerformanceCounter(&base);
#endif

#ifdef __unix__
	getcwd(launchDir, _MAX_PATH);
	strcat(launchDir, "/");
#else
	strcpy(launchDir, argv[0]);
	for(i = strlen(launchDir)-1; i>0 ; i--) {
		if((launchDir[i-1] == '/') || (launchDir[i-1] == '\\')) {
			launchDir[i] = '\0';
			break;
		}
	}
#endif

	strcpy(save_path, launchDir);
	strcat(save_path, "save/");

#ifndef WIN32
	mkdir (save_path, 0777);
	chown (save_path, getuid (), getgid ());
#endif

	strcpy(LastPath, launchDir);
	strcat(LastPath, "roms/");

	strcpy(config_file, save_path);
	strcat(config_file, "neocdgp2x.cfg");

	sound_set_volume(50);

	fp = fopen(config_file, "r");
	if(fp) {
		fscanf(fp, "%d", &skin);
		fclose(fp);
	} else
		skin = 0;
	if((skin < 0) || (skin > 9)) skin = 0;

	now_loading();

#ifndef GP2X
	//set window title
	SDL_WM_SetCaption(neocd_wm_title,neocd_wm_title);

	//set up window icon
	iconfp=SDL_RWFromMem((void*)rawicon, sizeof(rawicon));
	icon=SDL_LoadBMP_RW(iconfp,0);
	SDL_WM_SetIcon(icon,NULL);
#endif

	// Displays version number, date and time
	puts(VERSION1);
	puts(VERSION2);

	// Allocate needed memory buffers
	puts("NEOGEO: Allocating memory...");
	if (memory_init() == 0) {
		printf("Fatal Error: Could not load NEOCD.BIN\n");
		errorMsg("Fatal Error: Could not load NEOCD.BIN\n");
		return 0;
	}

	// System init
	sound_enable = sound_init();
	neogeo_video_init();
	input_init();
	printf("System initialized%s.\n", sound_enable ? "" : "(Sound disable)");

	while(1) {
		// Initialize End... (Filer Start)
		reset_joystick(0);
		if(!(filer_option = filer(launchDir, LastPath, game_dir, &skin))) {
			exit(0);
		} else {
			fp = fopen(config_file, "w");
			if(fp) {
				fprintf(fp, "%d\n", skin);
				fclose(fp);
			}
		}
		//printf("Filer Option: %d.\n", filer_option);
		//filer_option = 1;
		now_loading();

		if (neogeo_init())
			neogeo_run();

		break;
	}

	return 0;
}

//----------------------------------------------------------------------------
int neogeo_init(void)
{
	memset(memory_region_cpu1, 0, 0x200000);
	memset(memory_region_cpu2, 0, 0x10000);

	neogeo_driver_init();
	input_reset();
	cdrom_init();
	cdda_init();
	YM2610_sh_start();
	printf("Initialize everything.\n");

	cpu_reset_flag = 0;
	running = 1;

	swab(startup_bin, memory_region_cpu1 + 0x10f300, 0xd00);
	m68000_write_memory_32(0x120002, 0xffffffff);
	m68000_write_memory_32(0x11c808, 0xc0c760);	// load screen setup function
	m68000_write_memory_32(0x11c80c, 0xc0c814);	// load screen progress function
	m68000_write_memory_32(0x11c810, 0xc190e2);	// load screen default anime data

	game_index = 99;
	if (!neogeo_boot_bios)
	{
		int i;

		if (!cdrom_process_ipl())
		{
			printf("Error: Error while processing IPL.TXT.\n");
			return 0;
		}

		neogeo_read_gamename();
		neogeo_ngh = m68000_read_memory_16(0x108);

		for (i = 0; games[i].ngh_number; i++)
		{
			if (games[i].ngh_number == neogeo_ngh)
			{
				game_index = i + 1;
				strcpy(game_name, games[i].name);

				break;
			}
		}

		if (NGH_NUMBER(0x0085))
		{
			u16 *mem16 = (u16 *)memory_region_cpu1;

			mem16[0x132020 >> 1] = 0x4ef9;
			mem16[0x132022 >> 1] = 0x00c0;
			mem16[0x132024 >> 1] = 0xdb60;

			mem16[0x132026 >> 1] = 0x4ef9;
			mem16[0x132028 >> 1] = 0x00c0;
			mem16[0x13202a >> 1] = 0xdb6a;

			strcpy(config_game_name, "SAMURAI SPIRITS RPG");
		}

		printf("NeoGeo NGH Number: 0x%04x\n", neogeo_ngh);
		printf("NeoGeo Game Title: %s\n", config_game_name);
		printf("NeoGeo Game Index: %d\n", game_index);
		printf("NeoGeo Game Name: %s\n", game_name);
	}

	return running;
}

//----------------------------------------------------------------------------
void neogeo_reset(void)
{
	timer_init(0);
	input_reset();
	cdda_stop();
	frameskip_reset();

	neogeo_driver_reset();
	neogeo_video_reset();

	if (game_index != 99)
	{
		if (cpu_reset_flag == 0)
		{
			m68000_set_reg(M68K_PC, 0xc0a822);

			m68000_write_memory_32(0x108000 + 0x76ee, m68000_read_memory_32(0x68));	// IRQ vector
//			m68000_write_memory_32(0x108000 + 0x76f2, m68000_read_memory_32(0x64));	// VBLANK

			m68000_write_memory_8(0x108000 + 0x76d9, 0x01);
			m68000_write_memory_8(0x108000 + 0x76c3, 0x01);
			m68000_write_memory_8(0x108000 + 0x764b, 0xff);
			m68000_write_memory_8(0x108000 + 0x76f6, 0xff);
			m68000_write_memory_8(0x108000 + 0x7e85, 0x01);
		}
		else
		{
			m68000_set_reg(M68K_PC, 0x000122);

			m68000_write_memory_8(0x108000 + 0x7d80, 0x82);
			m68000_write_memory_8(0x108000 + 0x7daf, 0x01);
			m68000_write_memory_8(0x108000 + 0x7ee1, 0x0a);
			m68000_write_memory_8(0x108000 + 0x7675, 0x01);
			m68000_write_memory_8(0x108000 + 0x7ebf, 0x00);
			m68000_write_memory_32(0x108000 + 0x7db6, 0);
			m68000_write_memory_32(0x108000 + 0x7dba, 0);
		}

		cpu_reset_flag = 1;
	}

	m68000_write_memory_16(0xff011c, ~(neogeo_region << 8));
	m68000_write_memory_8(0x10fd83, neogeo_region);
}

//----------------------------------------------------------------------------
void neogeo_shutdown(void)
{
	// Close everything and free memory
	sound_pause(1);
	YM2610_sh_stop();
	cdda_shutdown();
	sound_shutdown();
	cdrom_shutdown();
	input_shutdown();
	video_shutdown();
	memory_shutdown();

	printf("NEOGEO: System Shutdown.\n");

	if(loading)
		SDL_FreeSurface(loading);

	SDL_Quit();

#ifdef GP2X
	chdir("/usr/gp2x");
	execl("/usr/gp2x/gp2xmenu", "/usr/gp2x/gp2xmenu", NULL);
#endif

	return;
}

//----------------------------------------------------------------------------
void neogeo_exit(void)
{
	puts("NEOGEO: Exit requested by software...");
	exit(0);
}

//----------------------------------------------------------------------------
void neogeo_run(void)
{
	Uint32 now;
	int i;
	int z80_timeslice;
	int m68k_timeslice;
	int skip_count;

	neogeo_raster_enable = 0;

	neogeo_reset();
	sound_pause(0);

	z80_timeslice = 73333 / RASTER_LINES;
	m68k_timeslice = 200000 / RASTER_LINES;
	skip_next_frame = 0;
	frame_count = 0;

	while(running)
	{
		for (scanline = 1; scanline <= RASTER_LINES; scanline++) {
	    	m68k_execute(m68k_timeslice);
			z80_exec(z80_timeslice);
			my_timer();
			if(neogeo_driver_type != NORMAL)
				neogeo_raster_interrupt(scanline, neogeo_driver_type);
		}

		if(neogeo_driver_type == NORMAL)
			neogeo_interrupt();
   		neogeo_cdda_check();

		/* check the watchdog */
		if (watchdog_counter > 0) {
			if (--watchdog_counter == 0) {
				//logerror("reset caused by the watchdog\n");
				printf("reset caused by the watchdog\n");
				neogeo_reset();
			}
		}

		video_draw_screen(RASTER_LINES);

		if (neogeo_loadfinished)
		{
			neogeo_loadfinished = 0;
			video_enable_w(1);
			fix_disable_w(0);
			spr_disable_w(0);
		}

		update_inputport();

		frame_count++;
		if(!skip_next_frame) draw_count++;

		now = get_ticks();
		if(now > next_time) {
			printf("FPS: %.3f/%.3f\n", draw_count / 8.0, frame_count / 8.0);
			next_time += (ONESEC << 3);
			frame_count = draw_count = 0;
		}

		/* Speed Throttle */
#if 1 // frame limit...
		if ((filer_option < 4) && (now < neocd_time)) {
			skip_next_frame = 0;
            while(get_ticks() < neocd_time) {
#ifdef WIN32
                if((neocd_time - get_ticks()) > 20000)
                    Sleep(20);
                else
                    Sleep(0);
#endif
			}
		}
#if 1 // frame skip
		else if(filer_option < 3){
			skip_next_frame = 1;
			skip_count++;
			if(skip_count > 5) {
				skip_next_frame = 0;
				skip_count = 0;
			}
		}
#endif
#endif
		neocd_time += REFRESHTIME;
	}

	return;
}

//----------------------------------------------------------------------------
void neogeo_read_gamename(void)
{
	u8 *ptr;
	int temp;

	ptr = memory_region_cpu1 + m68000_read_memory_32(0x11A);
	swab(ptr, config_game_name, 80);

	for(temp=0;temp<80;temp++) {
		if (!isprint(config_game_name[temp])) {
			config_game_name[temp]=0;
			break;
		}
	}
}

//----------------------------------------------------------------------------
TICKER get_ticks(void)
{
#ifdef GP2X
    struct timeval tv;
    gettimeofday(&tv, 0);
    return (TICKER)tv.tv_sec * TICKSPERSEC + tv.tv_usec;
#else
    LARGE_INTEGER tim;
    QueryPerformanceCounter(&tim);
    return (TICKER)((tim.QuadPart - base.QuadPart) * TICKSPERSEC / freq.QuadPart);
#endif
}

//----------------------------------------------------------------------------
void frameskip_reset(void)
{
    frame_count = draw_count = 0;
    neocd_time = get_ticks();
	next_time = neocd_time + (ONESEC << 3);
	neocd_time += REFRESHTIME;
}

//----------------------------------------------------------------------------
void open_joystick(void)
{
#ifdef GP2X
	int i;

	while(joy_count)
		joy_close(joys[--joy_count]);

	for (i=1; i<16; i++) {
		struct usbjoy *joy = joy_open(i);
		if(joy != NULL) {
			int j = 0;
			const char *name = joy_name(joy);

			while(joy_list[j].name) {
				if(strcmp(joy_list[j].name, name) == 0) break;
				j++;
			}
			joy_type[joy_count] = joy_list[j].type;
			joys[joy_count++] = joy;

			if(joy_count == 2) break;
		}
	}
#else
	joy_count = 0;
#endif

	msg_printf("%d USB device found.\n", joy_count);
}

u32 reset_joystick(int num)
{
#ifdef GP2X
	if(num == 0) {
		int i;

		for(i = 0; i < GPX_VK_MAX; i++) input_map[i] = -1;

		input_map[GPX_VK_START] = P1_START;
		input_map[GPX_VK_SELECT] = P1_SELECT;

		input_map[GPX_VK_FX] = P1_BUTTONA;
		input_map[GPX_VK_FB] = P1_BUTTONB;
		input_map[GPX_VK_FA] = P1_BUTTONC;
		input_map[GPX_VK_FY] = P1_BUTTOND;

		input_map[GPX_VK_UP] = P1_UP;
		input_map[GPX_VK_DOWN] = P1_DOWN;
		input_map[GPX_VK_LEFT] = P1_LEFT;
		input_map[GPX_VK_RIGHT] = P1_RIGHT;

		input_map[GPX_VK_FX] = P1_BUTTONA;
		input_map[GPX_VK_FB] = P1_BUTTONB;
		input_map[GPX_VK_FA] = P1_BUTTONC;
		input_map[GPX_VK_FY] = P1_BUTTOND;
	}

	if(num < joy_count) {
		int i;
		for (i = 0; i < JOY_MAX; i++) joy_map[num][i] = -1;

		joy_map[num][JOY_UP] = P1_UP;
		joy_map[num][JOY_DOWN] = P1_DOWN;
		joy_map[num][JOY_LEFT] = P1_LEFT;
		joy_map[num][JOY_RIGHT] = P1_RIGHT;

		switch(joy_type[num]) {
			case JOY_GREEN:
				msg_printf("Joystick %d: GreenAsia type\n", num);
				joy_map[num][JOY_BUTTON + 5] = P1_SELECT;
				joy_map[num][JOY_BUTTON + 6] = P1_START;

				joy_map[num][JOY_BUTTON + 9] = P1_SELECT;
				joy_map[num][JOY_BUTTON + 10] = P1_START;
				joy_map[num][JOY_BUTTON + 11] = P1_START;
				joy_map[num][JOY_BUTTON + 12] = P1_START;

				joy_map[num][JOY_BUTTON + 3] = P1_BUTTONA;
				joy_map[num][JOY_BUTTON + 4] = P1_BUTTONB;
				joy_map[num][JOY_BUTTON + 1] = P1_BUTTONC;
				joy_map[num][JOY_BUTTON + 2] = P1_BUTTOND;
				break;
			case JOY_DRAGON:
				msg_printf("Joystick %d: DragonRise type\n", num);
				joy_map[num][JOY_BUTTON + 7] = P1_SELECT;
				joy_map[num][JOY_BUTTON + 9] = P1_START;

				joy_map[num][JOY_BUTTON + 11] = P1_START;
				joy_map[num][JOY_BUTTON + 12] = P1_START;

				joy_map[num][JOY_BUTTON + 3] = P1_BUTTONA;
				joy_map[num][JOY_BUTTON + 4] = P1_BUTTONB;
				joy_map[num][JOY_BUTTON + 1] = P1_BUTTONC;
				joy_map[num][JOY_BUTTON + 2] = P1_BUTTOND;
				break;
			case JOY_BASIC:
			default:
				msg_printf("Joystick %d: Basic or Unknown type\n", num); // default set from mamegp2x
				joy_map[num][JOY_BUTTON + 1] = P1_BUTTOND;
				joy_map[num][JOY_BUTTON + 2] = P1_BUTTONB;
				joy_map[num][JOY_BUTTON + 3] = P1_BUTTONA;
				joy_map[num][JOY_BUTTON + 4] = P1_BUTTONC;

				joy_map[num][JOY_BUTTON + 9] = P1_SELECT;
				joy_map[num][JOY_BUTTON + 10] = P1_START;
				joy_map[num][JOY_BUTTON + 11] = P1_START;
				joy_map[num][JOY_BUTTON + 12] = P1_START;
				break;
		}
	}
#endif
}

u32 read_joystick(int num)
{
	u32 buttons = 0;
	int i;

#ifdef GP2X
	if (num == 0) {
		#define PRESS(x) (gp2xbtn & BTN(x))
		static u32 prev = 0;
		u32 gp2xbtn;
		u32 change = 0;
		int fn = 0;

		if((joy_fd < 0) || (read(joy_fd, &gp2xbtn, 4) != 4)) gp2xbtn = 0;

		change = prev ^ gp2xbtn;
		prev = gp2xbtn;

		if (gp2xbtn & GPX_CROSS_MASK) gp2xbtn &= ~GPX_DIAG_MASK;
		if (PRESS(GPX_VK_UP_LEFT)) gp2xbtn |= BTN(GPX_VK_UP) | BTN(GPX_VK_LEFT);
		if (PRESS(GPX_VK_UP_RIGHT)) gp2xbtn |= BTN(GPX_VK_UP) | BTN(GPX_VK_RIGHT);
		if (PRESS(GPX_VK_DOWN_LEFT)) gp2xbtn |= BTN(GPX_VK_DOWN) | BTN(GPX_VK_LEFT);
		if (PRESS(GPX_VK_DOWN_RIGHT)) gp2xbtn |= BTN(GPX_VK_DOWN) | BTN(GPX_VK_RIGHT);

		if (PRESS(GPX_VK_SELECT) && ((PRESS(GPX_VK_FL) && PRESS(GPX_VK_FR)) || PRESS(GPX_VK_START))) {
			//menu_mode = 1;
			running = 0;
			gp2xbtn &= ~(BTN(GPX_VK_SELECT) | BTN(GPX_VK_FL) | BTN(GPX_VK_FR) | BTN(GPX_VK_START));
		}

		if ((change & BTN(GPX_VK_START)) && PRESS(GPX_VK_FL) && PRESS(GPX_VK_FR)) {
			if(PRESS(GPX_VK_START))
				buttons |= BTN(P1_START);
			else
				buttons &= ~BTN(P1_START);
			gp2xbtn &= ~(BTN(GPX_VK_START) | BTN(GPX_VK_FL) | BTN(GPX_VK_FR));
		}

		if (gp2xbtn & change & BTN(GPX_VK_VOL_UP)) {
			if (PRESS(GPX_VK_SELECT)) {
				/*
				if(option_autoframeskip) {
					option_autoframeskip = 0;
					option_frameskip = 0;
					autoframeskip_reset();
					msg_printf("Frame skip: %d\n", option_frameskip);
				} else if(option_frameskip < 11) {
					++option_frameskip;
					autoframeskip_reset();
					msg_printf("Frame skip: %d\n", option_frameskip);
				}
				*/
			} else {
				if(sound_volume < 30)
					sound_set_volume(sound_volume + 2);
				else
					sound_set_volume(sound_volume + 5);
			}
		}

		if (gp2xbtn & change & BTN(GPX_VK_VOL_DOWN)) {
			if (PRESS(GPX_VK_SELECT)) {
				/*
				option_autoframeskip = 0;
				if(option_frameskip > 0) {
					--option_frameskip;
					autoframeskip_reset();
					msg_printf("Frame skip: %d\n", option_frameskip);
				} else {
					option_autoframeskip = 1;
					autoframeskip_reset();
					msg_printf("Frame skip: Auto\n");
				}
				*/
			} else {
				if(sound_volume > 30)
					sound_set_volume(sound_volume - 5);
				else
					sound_set_volume(sound_volume - 2);
			}
		}

		for(i = 0; i < GPX_VK_MAX; i++) {
			if((input_map[i] != -1) && PRESS(i))
				buttons |= BTN(input_map[i]);
		}

		#undef PRESS
	}

	if (num < joy_count) {
		buttons = buttons & (BTN(P1_SELECT) | BTN(P1_START));
		joy_update(joys[num]);
		if(joy_getaxe(UP, joys[num])) buttons |= BTN(joy_map[num][JOY_UP]);
		if(joy_getaxe(DOWN, joys[num])) buttons |= BTN(joy_map[num][JOY_DOWN]);
		if(joy_getaxe(LEFT, joys[num])) buttons |= BTN(joy_map[num][JOY_LEFT]);
		if(joy_getaxe(RIGHT, joys[num])) buttons |= BTN(joy_map[num][JOY_RIGHT]);

		for (i = 0; i < 32; i++) {
			if((joy_map[num][JOY_BUTTON1 + i] != -1) && joy_getbutton(i, joys[num]))
				buttons |= BTN(joy_map[num][JOY_BUTTON1 + i]);
		}
	}
#else
	SDL_Event event;
	static int pressed = 0;
	int button = -1;

	if(num != 0) return 0;

	SDL_Delay(1);
	while(SDL_PollEvent(&event)) {
		switch(event.type) {
			case SDL_QUIT:
				running = 0; //Loop = LOOP_EXIT;
				break;
			case SDL_KEYDOWN:
				switch(event.key.keysym.sym) {
					case SDLK_ESCAPE: running = 0; break;
					//case SDLK_F5: Loop = LOOP_RESET; break;
					//case SDLK_F12: menu_mode = 1; break;
				}
			case SDL_KEYUP:
				button = -1;
				switch(event.key.keysym.sym) {
					case SDLK_1: button = P1_START; break;
					case SDLK_2: button = P1_SELECT; break;
					case 'z': button = P1_BUTTONA; break;
					case 'x': button = P1_BUTTONB; break;
					case 'a': button = P1_BUTTONC; break;
					case 's': button = P1_BUTTOND; break;
					case 'q': button = P1_AF_A; break;
					case 'w': button = P1_AF_B; break;
					case 'e': button = P1_AF_C; break;
					case 'r': button = P1_AF_D; break;
					case SDLK_LEFT: button = P1_LEFT; break;
					case SDLK_RIGHT: button = P1_RIGHT; break;
					case SDLK_UP: button = P1_UP; break;
					case SDLK_DOWN: button = P1_DOWN; break;
				}
				if(button != -1)
					if(event.type == SDL_KEYDOWN)
						pressed |= BTN(button);
					else
						pressed &= ~BTN(button);
				break;
		}
	}

	buttons = pressed;
#endif

	return buttons;
}
